@@ -1,21 +1,30 @@
-use crate::Error;
+use crate::{Error, Result};
 use path_slash::PathBufExt;
 use std::path::PathBuf;
 use url::Url;
 
-/// A parsed href.
+/// An href can be an absolute url, an absolute path, or a relative path.
 ///
-/// Hrefs are used throughout the STAC specification to link between objects and
-/// to assets. They are defined in the specification as URIs, meaning they
-/// should always be `/`-delimited paths. The `Href` enum provides a
-/// platform-independent way to store and manipulate the paths.
+/// Hrefs are used throughout the STAC specification to link between objects, to
+/// assets, and to outside resources. They are defined in the specification as
+/// URIs, meaning they should always be `/`-delimited paths. This `Href` enum
+/// provides a platform-independent way to store and manipulate the paths.
 ///
-/// `Href`s are always created from `/`-delimited strings:
+/// `Href`s are always created from `/`-delimited strings. If you might be
+/// working with a `\` delimited string (e.g. on Windows), use [PathBufHref].
 ///
 /// ```
 /// use stac::Href;
 /// let path_href = Href::new("a/path/to/an/item.json");
 /// let url_href = Href::new("http://example.com/item.json");
+///
+/// #[cfg(target_os = "windows")]
+/// {
+///     use stac::PathBufHref;
+///     let path_buf_href = PathBufHref::new(r"a\path\to\an\item.json");
+///     let href = Href::from(path_buf_href);
+///     assert_eq!(href.as_str(), "a/path/to/an/item.json");
+/// }
 /// ```
 #[derive(Debug, PartialEq, Clone, Eq, Hash)]
 pub enum Href {
@@ -54,6 +63,8 @@ pub enum PathBufHref {
 impl Href {
     /// Creates an href.
     ///
+    /// Does not do any `\` conversion. If you need to handle possibly-`\`-delimited paths, use [PathBufHref].
+    ///
     /// # Examples
     ///
     /// ```
@@ -65,60 +76,54 @@ impl Href {
     /// assert!(href.is_url());
     /// assert_eq!(href.as_str(), "http://example.com/data");
     /// ```
-    pub fn new<S: ToString>(href: S) -> Href {
+    pub fn new(href: impl ToString) -> Href {
         let href = href.to_string();
-        match Url::parse(&href) {
-            Ok(url) => {
-                if url.cannot_be_a_base() {
-                    Href::Path(href)
-                } else {
-                    Href::Url(url)
-                }
+        if let Ok(url) = Url::parse(&href) {
+            if url.cannot_be_a_base() {
+                Href::Path(href)
+            } else {
+                Href::Url(url)
             }
-            Err(_) => Href::Path(href),
+        } else {
+            Href::Path(href)
         }
     }
 
     /// Joins this href to another href.
     ///
-    /// If the provided href is an absolute path or a url, just returns.
-    /// Otherwise, build a new path/url with the provided href.  If the base url
-    /// ends in a `/` it is unmodified, otherwise the last segment is treated as
-    /// a file name and dropped.
+    /// If the passed href is an absolute path or a url, this method just
+    /// returns the passed href as is.  Otherwise, it builds a new path or url
+    /// by joining this href and the provided href.  If `self` ends in a `/` it
+    /// is unmodified, otherwise the last segment is treated as a file name and
+    /// dropped.
     ///
     /// # Examples
     ///
     /// ```
     /// # use stac::Href;
     /// let base = Href::new("data/catalog.json");
-    /// let item = base.join("./collection/collection.json").unwrap();
-    /// assert_eq!(item.as_str(), "data/collection/collection.json");
+    /// let item = base.join("./item/item.json").unwrap();
+    /// assert_eq!(item.as_str(), "data/item/item.json");
+    /// let absolute_item = base.join("http://example.com/data/item.json").unwrap();
+    /// assert_eq!(absolute_item.as_str(), "http://example.com/data/item.json");
     ///
     /// let base = Href::new("data/");
-    /// let item = base.join("./collection/collection.json").unwrap();
-    /// assert_eq!(item.as_str(), "data/collection/collection.json");
+    /// let item = base.join("./item/item.json").unwrap();
+    /// assert_eq!(item.as_str(), "data/item/item.json");
     /// ```
-    pub fn join<T>(&self, href: T) -> Result<Href, Error>
-    where
-        T: Into<Href>,
-    {
+    pub fn join(&self, href: impl Into<Href>) -> Result<Href> {
         let href = href.into();
-        if href.is_url() || href.is_absolute_path() {
+        if href.is_absolute() {
             return Ok(href);
         }
         match self {
-            Href::Url(base) => {
-                let href = base.join(href.as_str()).map(Href::from)?;
-                Ok(href)
-            }
+            Href::Url(base) => base.join(href.as_str()).map(Href::Url).map_err(Error::from),
             Href::Path(base) => {
-                // Inspired by/taken from the url crate
-                let last_slash_index = base.rfind('/').unwrap_or(0);
-                let (directory, _) = base.split_at(last_slash_index);
-                let path = if directory.is_empty() {
+                let (path, _) = extract_path_filename(base);
+                let path = if path.is_empty() {
                     href.into_string()
                 } else {
-                    format!("{}/{}", directory, href.as_str())
+                    format!("{}/{}", path, href.as_str())
                 };
                 let path = normalize_path(path);
                 Ok(Href::Path(path))
@@ -126,7 +131,7 @@ impl Href {
         }
     }
 
-    /// Returns true if this href is a url.
+    /// Returns `true` if this href is a url.
     ///
     /// # Examples
     ///
@@ -161,7 +166,7 @@ impl Href {
         }
     }
 
-    /// Returns true if this href is a path.
+    /// Returns `true` if this href is a path.
     ///
     /// # Examples
     ///
@@ -176,7 +181,7 @@ impl Href {
         matches!(self, Href::Path(_))
     }
 
-    /// Returns a reference to this href as a str.
+    /// Returns this href as a [&str].
     ///
     /// # Examples
     ///
@@ -192,9 +197,9 @@ impl Href {
         }
     }
 
-    /// Returns true if this is an absolute href.
+    /// Returns `true` if this is an absolute href.
     ///
-    /// Urls are always absolute.
+    /// [Href::Urls](Href::Url) are always absolute.
     ///
     /// # Examples
     ///
@@ -205,74 +210,46 @@ impl Href {
     /// assert!(!Href::new("a/relative/path").is_absolute());
     /// ```
     pub fn is_absolute(&self) -> bool {
-        self.is_absolute_path() || self.is_url()
+        match self {
+            Href::Url(_) => true,
+            Href::Path(path) => is_absolute(path),
+        }
     }
 
-    /// Ensures that this href ends in a slash.
+    /// Converts this href into an absolute one.
+    ///
+    /// For [Href::Url], this is a noop.  For [Href::Path], this will return an
+    /// error if [std::fs::canonicalize] fails.
     ///
     /// # Examples
     ///
     /// ```
     /// # use stac::Href;
     /// let mut href = Href::new("http://example.com/data");
-    /// href.ensure_ends_in_slash();
-    /// assert_eq!(href.as_str(), "http://example.com/data/");
-    /// href.ensure_ends_in_slash();
-    /// assert_eq!(href.as_str(), "http://example.com/data/");
+    /// href.make_absolute().unwrap();
+    /// assert_eq!(href.as_str(), "http://example.com/data");
     ///
-    ///
-    /// let mut href = Href::new("data");
-    /// href.ensure_ends_in_slash();
-    /// assert_eq!(href.as_str(), "data/");
-    /// href.ensure_ends_in_slash();
-    /// assert_eq!(href.as_str(), "data/");
-    /// ```
-    pub fn ensure_ends_in_slash(&mut self) {
-        match self {
-            Href::Url(url) => {
-                if !url.path().ends_with('/') {
-                    let _ = url
-                        .path_segments_mut()
-                        .expect("the url should have a path")
-                        .push("");
-                }
-            }
-            Href::Path(path) => {
-                if !path.ends_with('/') {
-                    path.push('/');
-                }
-            }
-        }
-    }
-
-    /// Converts this href into an absolute one.
-    ///
-    /// Will return an error if [std::fs::canonicalize] fails.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// # use stac::Href;
-    /// let href = Href::new("http://example.com/data");
-    /// assert_eq!(href.make_absolute().unwrap().as_str(), "http://example.com/data");
-    ///
-    /// let href = Href::new("data/catalog.json").make_absolute().unwrap();
+    /// let mut href = Href::new("data/catalog.json");
+    /// href.make_absolute().unwrap();
     /// let err = Href::new("not/a/real/path").make_absolute().unwrap_err();
     ///
-    pub fn make_absolute(self) -> Result<Href, Error> {
-        let href = PathBufHref::from(self);
-        match href {
-            PathBufHref::Url(_) => Ok(href.into()),
-            PathBufHref::Path(path) => {
+    pub fn make_absolute(&mut self) -> Result<()> {
+        if let Href::Path(path) = self {
+            if let PathBufHref::Path(path) = PathBufHref::from(path.as_str()) {
                 let path = std::fs::canonicalize(path)?;
-                Ok(PathBufHref::Path(path).into())
+                *self = PathBufHref::Path(path).into();
             }
         }
+        Ok(())
     }
 
-    /// Converts the provided link into a relative one.
+    /// Converts the provided href into a relative href, relative to `self`.
     ///
-    /// Adds an `"./"` to the front of "downward" hrefs.
+    /// - Adds an `"./"` to the front of "downward" hrefs.
+    /// - If both paths don't share a common base, they are assumed to be in the
+    /// same parent directory.
+    /// - If both paths are absolute and they do not share a common base,
+    /// returns the original href unchanged.
     ///
     /// # Examples
     ///
@@ -281,40 +258,52 @@ impl Href {
     /// let catalog = Href::new("data/catalog.json");
     /// let collection = Href::new("data/collection/collection.json");
     /// assert_eq!(
-    ///     catalog.make_relative(&collection).as_str(),
+    ///     catalog.make_relative(collection.clone()).as_str(),
     ///     "./collection/collection.json"
     /// );
     /// assert_eq!(
-    ///     collection.make_relative(&catalog).as_str(),
+    ///     collection.make_relative(catalog).as_str(),
     ///     "../catalog.json"
     /// );
-    pub fn make_relative(&self, href: &Href) -> Href {
+    pub fn make_relative(&self, href: Href) -> Href {
         match self {
             Href::Url(base) => match href {
-                Href::Url(url) => {
-                    if let Some(path) = base.make_relative(url) {
+                Href::Url(url) => base
+                    .make_relative(&url)
+                    .map(|path| {
                         if path.is_empty() {
                             let (_, file_name) = extract_path_filename(url.path());
                             Href::Path(format!("./{}", file_name))
                         } else {
                             Href::Path(path)
                         }
+                    })
+                    .unwrap_or_else(|| Href::Url(url)),
+                // We skip the leading slash on the path to get make relative to go.
+                Href::Path(path) => {
+                    if is_absolute(&path) {
+                        Href::Path(path)
                     } else {
-                        // TODO test and implement
-                        unimplemented!()
+                        Href::Path(make_relative(&base.path()[1..], &path))
                     }
                 }
-                // We skip the leading slash on the path to get make relative to go.
-                Href::Path(path) => Href::Path(make_relative(&base.path()[1..], path)),
             },
             Href::Path(base) => match href {
-                Href::Url(url) => Href::Url(url.clone()),
-                Href::Path(path) => Href::Path(make_relative(base, path)),
+                Href::Url(url) => Href::Url(url),
+                Href::Path(path) => {
+                    if is_absolute(base) && is_absolute(&path) {
+                        Href::Path(path)
+                    } else {
+                        Href::Path(make_relative(base, &path))
+                    }
+                }
             },
         }
     }
 
-    /// Rebases a href from one root to another.
+    /// Rebases a relative href from one root to another.
+    ///
+    /// If `self` is a url or absolute, this is a noop.
     ///
     /// # Examples
     ///
@@ -326,23 +315,50 @@ impl Href {
     /// item.rebase(&old_root_catalog, &new_root).unwrap();
     /// assert_eq!(item.as_str(), "a/new/base/item/item.json");
     /// ```
-    pub fn rebase(&mut self, from: &Href, to: &Href) -> Result<(), Error> {
-        *self = match self {
-            Href::Url(_) => return Ok(()),
-            Href::Path(path) => {
-                if is_absolute(path) {
-                    return Ok(());
-                }
-                to.join(make_relative(from.as_str(), path))?
+    pub fn rebase(&mut self, from: &Href, to: &Href) -> Result<()> {
+        if let Href::Path(path) = self {
+            if is_absolute(path) {
+                return Ok(());
             }
-        };
+            *self = to.join(make_relative(from.as_str(), path))?;
+        }
         Ok(())
     }
 
-    fn is_absolute_path(&self) -> bool {
+    /// Returns this href's file name.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use stac::Href;
+    /// let href = Href::new("a/path/item.json");
+    /// assert_eq!(href.file_name(), "item.json");
+    /// ```
+    pub fn file_name(&self) -> &str {
+        extract_path_filename(self.as_str()).1
+    }
+
+    /// Returns this href's directory.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use stac::Href;
+    /// let href = Href::new("a/path/item.json");
+    /// assert_eq!(href.directory(), "a/path");
+    /// ```
+    pub fn directory(&self) -> &str {
+        extract_path_filename(self.as_str()).0
+    }
+
+    pub(crate) fn ensure_ends_in_slash(&mut self) {
         match self {
-            Href::Path(path) => is_absolute(path),
-            _ => false,
+            Href::Url(url) => {
+                if let Ok(mut path_segments) = url.path_segments_mut() {
+                    let _ = path_segments.push("/");
+                }
+            }
+            Href::Path(path) => path.push('/'),
         }
     }
 
@@ -355,7 +371,30 @@ impl Href {
 }
 
 impl PathBufHref {
-    fn new<T: ToString>(href: T) -> PathBufHref {
+    /// Creates a new href with system-specific delimiters.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use stac::{PathBufHref, Href};
+    ///
+    /// #[cfg(target_os = "windows")]
+    /// {
+    ///     use stac::PathBufHref;
+    ///     let path_buf_href = PathBufHref::new(r"a\path\to\an\item.json");
+    ///     let href = Href::from(path_buf_href);
+    ///     assert_eq!(href.as_str(), "a/path/to/an/item.json");
+    /// }
+    ///
+    /// #[cfg(not(target_os = "windows"))]
+    /// {
+    ///     use stac::PathBufHref;
+    ///     let path_buf_href = PathBufHref::new("a/path/to/an/item.json");
+    ///     let href = Href::from(path_buf_href);
+    ///     assert_eq!(href.as_str(), "a/path/to/an/item.json");
+    /// }
+    /// ```
+    pub fn new(href: impl ToString) -> PathBufHref {
         Href::new(href).into()
     }
 }
@@ -414,6 +453,15 @@ impl From<String> for Href {
     }
 }
 
+impl From<Href> for String {
+    fn from(href: Href) -> String {
+        match href {
+            Href::Url(url) => url.to_string(),
+            Href::Path(path) => path,
+        }
+    }
+}
+
 impl From<&str> for PathBufHref {
     fn from(s: &str) -> PathBufHref {
         PathBufHref::new(s)
@@ -611,57 +659,60 @@ mod tests {
         assert_eq!(href.as_str(), "http://example.com/catalog.json");
     }
 
-    #[test]
-    fn ensure_path_ends_in_slash() {
-        let mut href = Href::new("data");
-        href.ensure_ends_in_slash();
-        assert_eq!(href.as_str(), "data/");
-    }
-
-    #[test]
-    fn ensure_url_ends_in_slash() {
-        let mut href = Href::new("http://example.com");
-        href.ensure_ends_in_slash();
-        assert_eq!(href.as_str(), "http://example.com/");
-        let mut href = Href::new("http://example.com/data");
-        href.ensure_ends_in_slash();
-        assert_eq!(href.as_str(), "http://example.com/data/");
-    }
-
     #[test]
     fn make_path_relative() {
         let base = Href::new("data/catalog.json");
         let target = Href::new("data/extensions-collection/collection.json");
         assert_eq!(
-            base.make_relative(&target).as_str(),
+            base.make_relative(target.clone()).as_str(),
             "./extensions-collection/collection.json"
         );
-        assert_eq!(target.make_relative(&base).as_str(), "../catalog.json");
-        assert_eq!(base.make_relative(&base).as_str(), "./catalog.json");
         assert_eq!(
-            base.make_relative(&Href::new("http://example.com/item.json"))
+            target.make_relative(base.clone()).as_str(),
+            "../catalog.json"
+        );
+        assert_eq!(base.make_relative(base.clone()).as_str(), "./catalog.json");
+        assert_eq!(
+            base.make_relative(Href::new("http://example.com/item.json"))
                 .as_str(),
             "http://example.com/item.json"
         );
     }
 
+    #[test]
+    fn make_path_relative_no_common_base() {
+        let base = Href::new("data/catalog.json");
+        let target = Href::new("other/extensions-collection/collection.json");
+        assert_eq!(
+            base.make_relative(target.clone()).as_str(),
+            "../other/extensions-collection/collection.json"
+        );
+
+        let base = Href::new("/data/catalog.json");
+        let target = Href::new("/other/extensions-collection/collection.json");
+        assert_eq!(base.make_relative(target.clone()), target);
+
+        let base = Href::new("http://example.com/catalog.json");
+        let target = Href::new("http://example.org/item/item.json");
+        assert_eq!(base.make_relative(target.clone()), target);
+    }
+
     #[test]
     fn make_url_relative() {
         let base = Href::new("http://example.com/data/catalog.json");
         let target = Href::new("data/extensions-collection/collection.json");
         assert_eq!(
-            base.make_relative(&target).as_str(),
+            base.make_relative(target.clone()).as_str(),
             "./extensions-collection/collection.json"
         );
-        assert_eq!(
-            target.make_relative(&base).as_str(),
-            "http://example.com/data/catalog.json"
-        );
-        assert_eq!(base.make_relative(&base).as_str(), "./catalog.json");
+        assert_eq!(base.make_relative(base.clone()).as_str(), "./catalog.json");
+
+        let target = Href::new("/data/extensions-collection/collection.json");
+        assert_eq!(base.make_relative(target.clone()), target);
     }
 
     #[test]
-    fn rebase() {
+    fn rebase_path_relative() {
         let old_root_catalog = Href::new("path/to/a/catalog.json");
         let new_root = Href::new("a/new/base/");
         let mut item = Href::new("path/to/a/item/item.json");
